Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Arg_parser.PrivateSourceLow-level functions intended for use within this library only.
A parser that just prints a usage message.
Replace a parser with a new parser that just prints the usage information about the original parser.
val finalize :
'a t ->
doc:string option ->
child_subcommands:Climate__.Subcommand.t list ->
prose:Manpage.Prose.t option ->
use_error_subcommand:bool ->
help_only_doc:string option ->
help_only_subcommands:Climate__.Subcommand.t list option ->
'a tval named_opt_for_internal :
?doc:string ->
?value_name:string ->
?hidden:bool ->
?completion:'a Completion.t ->
Climate__.Name.t Climate__.Climate_stdlib.Nonempty_list.t ->
'a conv ->
allow_many:bool ->
'a option tLike named_opt but takes its arguments as Name.ts rather than as strings is it's intended for use within this library.
val eval :
'a t ->
command_line:Climate__.Command_line.Rich.t ->
ignore_errors:bool ->
alt_subcommand_for_usage:string list option ->
alt_subcommand_for_errors:string list option ->
('a, Climate__.Non_ret.t) resultEvaluate a parser, parsing command-line arguments. The alt_subcommand_ arguments are to support commands like `program help subcommand` where we want usage strings to reference `program subcommand` and error messages to reference `program help`.