Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Climate_stdlib.OptionSourceinclude module type of OptionThe type for option values. Either None or a value Some v.
none is None.
some v is Some v.
value o ~default is v if o is Some v and default otherwise.
get o is v if o is Some v and raise otherwise.
join oo is Some v if oo is Some (Some v) and None otherwise.
product o0 o1 is Some (v0, v1) if o0 is Some v0 and o1 is Some v1 and None otherwise.
fold ~none ~some o is none if o is None and some v if o is Some v.
blend f o1 o2 is o1 if o2 is None, o2 if o1 is None, and Some (f v1 v2) if o1 is Some v1 and o2 is Some v2.
for_all p behaves like List.for_all p on a list of zero or one element:
for_all p None is true,for_all p (Some v) is p v.exists p behaves like List.exists p on a list of zero or one element:
exists p None is false,exists p (Some v) is p v.is_none o is true if and only if o is None.
is_some o is true if and only if o is Some o.
compare cmp o0 o1 is a total order on options using cmp to compare values wrapped by Some _. None is smaller than Some _ values.
to_result ~none o is Ok v if o is Some v and Error none otherwise.
to_list o is [] if o is None and [v] if o is Some v.
to_seq o is o as a sequence. None is the empty sequence and Some v is the singleton sequence containing v.